from typing import AsyncIterator
from rath.links.base import AsyncTerminatingLink
from rath.operation import GraphQLResult, Operation


class DirectSucceedingLink(AsyncTerminatingLink):
    """AssertLink is a link that asserts that the operation matches a set of assertions.

    This link is useful for testing.

    """

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """Executes an operation against the link

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Yields
        ------
        GraphQLResult
            The result of the operation
        """

        yield GraphQLResult(data={})



from types import TracebackType
from koil.composition import KoiledModel
from pydantic import Field, field_validator
from rath.errors import NotConnectedError
from rath.links.base import TerminatingLink, ContinuationLink
from typing import (
    AsyncGenerator,
    Dict,
    Any,
    Generator,
    Optional,
    Type,
)
from typing import Union
from graphql import (
    DocumentNode,
)
from rath.operation import GraphQLResult, opify
from contextvars import ContextVar, Token
from koil import unkoil_gen, unkoil


current_rath: ContextVar[Optional["Rath"]] = ContextVar("current_rath", default=None)


class Rath(KoiledModel):
    """A Rath is a client for a GraphQL API.

    Links are used to define the transport logic of the Rath. A Rath can be
    connected to a GraphQL API using a terminating link. By composing links to
    form a chain, you can define the transport logic of the client.

    For example, a Rath can be connected to a GraphQL API over HTTP by
    composing an HTTP terminating link with a RetryLink. The RetryLink will
    retry failed requests, and the HTTP terminating link will send the
    requests over HTTP.

    Example:
        ```python
        from rath import Rath
        from rath.links.retry import RetryLink
        from rathlinks.aiohttp import AIOHttpLink

        retry = RetryLink()
        http = AIOHttpLink(endpoint_url"https://graphql-pokemon.now.sh/graphql")

        rath = Rath(link=compose(retry, link))
        async with rath as r:
            await r.aquery(...)
        ```

    """

    link: TerminatingLink = Field(
        ...,
        description="The terminating link used to send operations to the server. Can be a composed link chain.",
    )
    """The terminating link used to send operations to the server. Can be a composed link chain."""

    _entered = False
    """An internal flag flag that indicates whether the Rath is currently in the context manager."""
    _context_token: Optional[Token[Optional["Rath"]]] = None
    """A context token that is used to keep track of the current rath in the context manager."""

    @field_validator("link", mode="before")
    def validate_link(
        cls, link: TerminatingLink | list[ContinuationLink | TerminatingLink] | Any
    ) -> TerminatingLink:
        """Validates the link and ensures it is a terminating link or a list of links."""
        if isinstance(link, list):
            # If the link is a list, we assume it is a chain of links
            # and we compose them into a single link
            from rath.links.compose import compose

            for lin in link:  # type: ignore
                if not isinstance(lin, (ContinuationLink, TerminatingLink)):
                    raise ValueError("All links in the list must be of type Link")

            if not isinstance(link[-1], TerminatingLink):
                raise ValueError("The last link in the list must be a TerminatingLink")

            return compose(*link)  # type: ignore

        if not isinstance(link, TerminatingLink):
            raise ValueError(
                "Link must be a TerminatingLink or a list of TerminatingLinks"
            )

        return link

    async def aquery(
        self,
        query: Union[str, DocumentNode],
        variables: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
        operation_name: Optional[str] = None,
        **kwargs: Any,
    ) -> GraphQLResult:
        """Query the GraphQL API.

        Takes a querystring, variables, and headers and returns the result.
        If provided, the operation_name will be used to identify which operation
        to execute.

        Args:
            query (str | DocumentNode): The query string or the DocumentNode.
            variables (Dict[str, Any], optional): The variables. Defaults to None.
            headers (Dict[str, Any], optional): Additional headers. Defaults to None.
            operation_name (str, optional): The operation_name to executed. Defaults to all.
            **kwargs: Additional arguments to pass to the link chain

        Raises:
            NotConnectedError: An error when the Rath is not connected and autoload is set to false

        Returns:
            GraphQLResult: The result of the query
        """
        op = opify(query, variables, headers, operation_name, **kwargs)

        result = None

        async for data in self.link.aexecute(op):
            result = data
            break

        if not result:
            raise NotConnectedError("Could not retrieve data from the server.")
            # This is to account for the fact that mypy apparently doesn't
            # understand that a return statement in a generator is valid.
            # This is a workaround to make mypy happy.

        return result

    def query(
        self,
        query: Union[str, DocumentNode],
        variables: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
        operation_name: Optional[str] = None,
        **kwargs: Any,
    ) -> GraphQLResult:
        """Query the GraphQL API.

        Takes a querystring, variables, and headers and returns the result.
        If provided, the operation_name will be used to identify which operation
        to execute.

        Args:
            query (str | DocumentNode): The query string or the DocumentNode.
            variables (Dict[str, Any], optional): The variables. Defaults to None.
            headers (Dict[str, Any], optional): Additional headers. Defaults to None.
            operation_name (str, optional): The operation_name to executed. Defaults to all.
            **kwargs: Additional arguments to pass to the link chain

        Raises:
            NotConnectedError: An error when the Rath is not connected and autoload is set to false

        Returns:
            GraphQLResult: The result of the query
        """
        return unkoil(self.aquery, query, variables, headers, operation_name)

    def subscribe(
        self,
        query: str,
        variables: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
        operation_name: Optional[str] = None,
        **kwargs: Any,
    ) -> Generator[GraphQLResult, None, None]:
        """Subscripe to a GraphQL API.

        Takes a querystring, variables, and headers and returns an async iterator
        that yields the results.

        Args:
            query (str | DocumentNode): The query string or the DocumentNode.
            variables (Dict[str, Any], optional): The variables. Defaults to None.
            headers (Dict[str, Any], optional): Additional headers. Defaults to None.
            operation_name (str, optional): The operation_name to executed. Defaults to all.
            **kwargs: Additional arguments to pass to the link chain

        Raises:
            NotConnectedError: An error when the Rath is not connected and autoload is set to false

        Yields:
            Iterator[GraphQLResult]: The result of the query as an async iterator
        """
        return unkoil_gen(
            self.asubscribe, query, variables, headers, operation_name, **kwargs
        )

    async def asubscribe(
        self,
        query: str,
        variables: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
        operation_name: Optional[str] = None,
        **kwargs: Any,
    ) -> AsyncGenerator[GraphQLResult, None]:
        """Subscripe to a GraphQL API.

        Takes a querystring, variables, and headers and returns an async iterator
        that yields the results.

        Args:
            query (str | DocumentNode): The query string or the DocumentNode.
            variables (Dict[str, Any], optional): The variables. Defaults to None.
            headers (Dict[str, Any], optional): Additional headers. Defaults to None.
            operation_name (str, optional): The operation_name to executed. Defaults to all.
            **kwargs: Additional arguments to pass to the link chain

        Raises:
            NotConnectedError: An error when the Rath is not connected and autoload is set to false

        Yields:
            Iterator[GraphQLResult]: The result of the query as an async iterator
        """

        op = opify(query, variables, headers, operation_name, **kwargs)
        async for data in self.link.aexecute(op):
            yield data

    async def __aenter__(self) -> "Rath":
        """Enters the context manager of the link"""
        self._entered = True
        self._context_token = current_rath.set(self)
        await self.link.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Exits the context manager of the link"""
        await self.link.__aexit__(exc_type, exc_val, traceback)
        self._entered = False
        if self._context_token:
            current_rath.set(None)



from http import HTTPStatus
import json
from typing import Any, Dict, List, Type, AsyncIterator
import httpx
from graphql import OperationType
from pydantic import Field
from rath.operation import GraphQLException, GraphQLResult, Operation
from rath.links.base import AsyncTerminatingLink
from rath.links.errors import AuthenticationError
import logging
from rath.links.types import Payload
from datetime import datetime

logger = logging.getLogger(__name__)


class DateTimeEncoder(json.JSONEncoder):
    """DateTimeEncoder is a JSONEncoder that extends the default python json decoder to serialize"""

    def default(self, o: Any):  # noqa
        """Override the default method to serialize datetime objects to ISO 8601 strings"""
        if isinstance(o, datetime):
            return o.isoformat()

        return json.JSONEncoder.default(self, o)


class HttpxLink(AsyncTerminatingLink):
    """HttpxLink is a terminating link that sends operations over HTTP using httpx"""

    endpoint_url: str
    """endpoint_url is the URL to send operations to."""

    auth_errors: List[HTTPStatus] = Field(
        default_factory=lambda: [
            HTTPStatus.FORBIDDEN,
        ]
    )
    """auth_errors is a list of HTTPStatus codes that indicate an authentication error."""
    json_encoder: Type[json.JSONEncoder] = Field(default=DateTimeEncoder, exclude=True)

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """Executes an operation against the link

        This link will create a new httpx session for each request. While
        we could also reuse the session, we currently don't do this. If
        you feel like this should be changed, please open an issue.

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Yields
        ------
        GraphQLResult
            The result of the operation
        """

        payload: Payload = {"query": operation.document}

        if operation.node.operation == OperationType.SUBSCRIPTION:
            raise NotImplementedError(
                "Aiohttp Transport does not support subscriptions"
            )

        if len(operation.context.files.items()) > 0:
            payload["variables"] = operation.variables

            files = operation.context.files

            file_map = {str(i): [path] for i, path in enumerate(files)}

            # Enumerate the file streams
            # Will generate something like {'0': <_io.BufferedReader ...>}
            file_streams = {str(i): files[path] for i, path in enumerate(files)}
            operations_str = json.dumps(payload, cls=self.json_encoder)
            file_map_str = json.dumps(file_map)

            post_kwargs: Dict[str, Any] = {
                "data": {
                    "operations": operations_str,
                    "map": file_map_str,
                },
                "files": file_streams,
            }

        else:
            payload["variables"] = operation.variables
            post_kwargs = {"json": payload}

        async with httpx.AsyncClient() as client:
            response = await client.post(
                self.endpoint_url, headers=operation.context.headers, **post_kwargs
            )

            if response.status_code in self.auth_errors:
                raise AuthenticationError(
                    f"Token Expired Error {operation.context.headers}"
                )

            if response.status_code == HTTPStatus.OK:
                json_response = response.json()

                if "errors" in json_response:
                    raise GraphQLException(
                        "\n".join([e["message"] for e in json_response["errors"]])
                    )

                if "data" not in json_response:
                    raise Exception(f"Response does not contain data {json_response}")

                yield GraphQLResult(data=json_response["data"])



from typing import AsyncIterator, Awaitable, Callable, Optional

from rath.links.base import ContinuationLink
from rath.operation import GraphQLResult, Operation
from rath.links.errors import AuthenticationError
from rath.errors import NotComposedError


class AuthTokenLink(ContinuationLink):
    """AuthTokenLink is a link that adds an authentication token to the context.
    The authentication token is retrieved by calling the token_loader function.
    If the wrapped link raises an AuthenticationError, the token_refresher function
    is called again to refresh the token.

    This link is statelss, and does not store the token. It is up to the user to
    store the token and pass it to the token_loader function.
    """

    maximum_refresh_attempts: int = 3
    """The maximum number of times the token_refresher function will be called, before the operation fails."""

    load_token_on_connect: bool = True
    """If True, the token_loader function will be called when the link is connected."""
    load_token_on_enter: bool = True
    """If True, the token_loader function will be called when the link is entered."""

    async def aload_token(self, operation: Operation) -> str:
        """A function that loads the authentication token.

        This function should return a string containing the authentication token.

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Raises
        ------
        Exception
            When the token cannot be loaded
        """
        raise Exception("No Token loader specified")

    async def arefresh_token(self, operation: Operation) -> str:
        """A function that refreshes the authentication token.

        This function should return a string containing the authentication token.
        In comparison to the token_loader function, this function is called when
        the server already raised an AuthenticationError, so a refresh should really
        be attempted.

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Raises
        ------
        Exception
            When the token cannot be refreshed
        """
        raise Exception("No Token refresher specified")

    async def aexecute(
        self, operation: Operation, retry: int = 0
    ) -> AsyncIterator[GraphQLResult]:
        """Executes and forwards an operation to the next link.

        This method will add the authentication token to the context of the operation,
        and will refresh the token if the next link raises an AuthenticationError, until
        the maximum number of refresh attempts is reached.

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Yields
        ------
        GraphQLResult
            The result of the operation
        """
        if not self.next:
            raise NotComposedError("No next link set")

        token = await self.aload_token(operation)
        operation.context.headers["Authorization"] = f"Bearer {token}"
        operation.context.initial_payload["token"] = token
        
        

        try:
            async for result in self.next.aexecute(operation):
                yield result

        except AuthenticationError as e:
            retry = retry + 1
            await self.arefresh_token(operation)
            if retry > self.maximum_refresh_attempts:
                raise AuthenticationError("Maximum refresh attempts reached") from e

            async for result in self.aexecute(operation, retry=retry + 1):
                yield result


class ComposedAuthLink(AuthTokenLink):
    """A composed version of the AuthTokenLink.

    This link is a composed link that allows to set the token_loader and token_refresher
    functions as composition elements not as class attributes.
    """

    token_loader: Optional[Callable[[], Awaitable[str]]]
    """The function used to load the authentication token. This function should
        return a string containing the authentication token."""
    token_refresher: Optional[Callable[[], Awaitable[str]]]
    """The function used to refresh the authentication token. This function should
        return a string containing the authentication token."""

    async def aload_token(self, operation: Operation) -> str:
        """Forwards the operation to the token_loader function."""
        if self.token_loader is None:
            raise Exception("No Token loader specified")
        return await self.token_loader()

    async def arefresh_token(self, operation: Operation) -> str:
        """Forwards the operation to the token_refresher function."""
        if self.token_refresher is None:
            raise Exception("No Token refresher specified")
        return await self.token_refresher()



from typing import Dict, Any, Callable, Optional


def recurse_parse_variables(
    variables: Dict[str, Any],
    predicate: Callable[[Optional[str], Any], bool],
    apply: Callable[[Any], Any],
) -> Dict[str, Any]:
    """Parse Variables

    Recursively traverse variables, applying the apply function to the value if the predicate
    returns True.

    Args:
        variables (Dict): The dictionary to parse.
        predicate (Callable[[str, Any], bool]):The path this is in
        apply (Callable[[Any], Any]): _description_

    Returns:
        Dict: _description_
    """

    def recurse_extract(obj: Any, path: Optional[str] = None) -> Any:
        """
        recursively traverse obj, doing a deepcopy, but
        replacing any file-like objects with nulls and
        shunting the originals off to the side.
        """

        if isinstance(obj, list):
            nulled_list = []
            for key, value in enumerate(obj):  # type: ignore
                value = recurse_extract(
                    value,
                    path=f"{path}.{key}" if path else str(key),
                )
                nulled_list.append(value)  # type: ignore
            return nulled_list  # type: ignore
        elif isinstance(obj, dict):
            nulled_obj = {}
            for key, value in obj.items():  # type: ignore
                value = recurse_extract(value, f"{path}.{key}" if path else str(key))  # type: ignore
                nulled_obj[key] = value
            return nulled_obj  # type: ignore
        elif predicate(path, obj):
            return apply(obj)
        else:
            return obj

    dicted_variables = recurse_extract(variables)

    return dicted_variables



from typing import AsyncIterator, Awaitable, Callable

from rath.links.base import ContinuationLink
from rath.operation import GraphQLResult, Operation
from rath.errors import NotComposedError


async def just_print(operation: Operation) -> None:
    """A simple logging function that prints the operation"""
    print(operation)


class LogLink(ContinuationLink):
    """The Log link is a logging
    link that logs the operation to the console.
    links.
    """

    log: Callable[[Operation], Awaitable[None]] = just_print

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """Executes an operation against the link

        This link simply forwards the operation to the next link.
        after logging it.

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Yields
        ------
        GraphQLResult
            The result of the operation
        """
        if not self.next:
            raise NotComposedError("No next link set")

        await self.log(operation)
        async for result in self.next.aexecute(operation):
            yield result



"""Turms related functions

Turms is a python library for generating Turms operations from python classes.
This module contains functions that allow you to execute these operations using
rath.

As well as a plugin that plays well with turms and autoconnects the generated
code to rath, allowing you to execute typesfave operations using rath.

"""



from pydantic import BaseModel
from typing import Any, Type, Protocol
from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema


class ID(str):
    """A custom scalar for IDs. If passed a pydantic model it an id property
    it will automatically validate to that models id"""

    def __get__(self, instance, owner) -> "ID": ...

    def __set__(self, instance, value: "IDCoercible") -> None: ...

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,  # noqa: ANN401
    ) -> CoreSchema:
        """Get the pydantic core schema for the interface"""
        return core_schema.no_info_before_validator_function(cls.validate, handler(str))

    @classmethod
    def validate(cls: Type["ID"], v: "IDCoercible") -> "ID":
        """Validate the ID"""
        if isinstance(v, BaseModel):
            if hasattr(v, "id"):
                return cls(v.id)  # type: ignore
            else:
                raise ValueError("It appears to be a pydantic model but has no id")

        if isinstance(v, str):
            return cls(v)

        if isinstance(v, int):
            return cls(str(v))

        raise TypeError(
            f"Needs to be either a instance of BaseModel (with an id) or a string got {type(v)}"
        )


class WithId(Protocol):
    "A protocol for objects that have an id attribute."

    id: ID


IDCoercible = str | ID | WithId | int
""" A type that can be coerced into an ID."""



from rath.rath import Rath, current_rath
from typing import AsyncIterator, Iterator, Protocol, Type, TypeVar, Any, Dict, Optional
from pydantic import BaseModel


# --- Base meta description ---
class TurmsMeta(Protocol):
    """Meta class for Turms operations"""

    document: str


TArgs = TypeVar("TArgs", bound=BaseModel)
TMeta = TypeVar("TMeta", bound=TurmsMeta)


class TurmsOperation(Protocol[TArgs, TMeta]):
    """Represents a Turms operation that is both callable and its own return type."""

    Meta: Type[TMeta]
    Arguments: Type[TArgs]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """A pydantic constructor"""
        ...


TOperation = TypeVar("TOperation", bound=TurmsOperation[Any, Any])


def execute(operation: Type[TOperation], variables: Dict[str, Any], rath: Optional[Rath] = None) -> TOperation:
    """Synchronously Executes an a query or mutation using rath


    This function will execute an operation using rath, retrieving
    the currentliy active rath client from the current_rath context
    if no rath client is provided.

    Parameters
    ----------
    operation : TurmsOperation
        The turms operation to execute
    variables : Dict[str, Any]
        The variables to use
    rath : Optional[Rath], optional
        The rath client, by default the current rath client

    Returns
    -------
    BaseModel
        The result of the operation
    """
    rath = rath or current_rath.get()
    assert rath is not None, "No rath client provided and no rath client in current_rath context"
    return operation(
        **rath.query(
            operation.Meta.document,
            operation.Arguments(**variables).model_dump(by_alias=True),
        ).data
    )


async def aexecute(operation: Type[TOperation], variables: Dict[str, Any], rath: Optional[Rath] = None) -> TOperation:
    """Asynchronously Executes a query or mutation using rath


    This function will execute an operation using rath, retrieving
    the currentliy active rath client from the current_rath context
    if no rath client is provided.

    Parameters
    ----------
    operation : TurmsOperation
        The turms operation to execute
    variables : Dict[str, Any]
        The variables to use
    rath : Optional[Rath], optional
        The rath client, by default the current rath client

    Returns
    -------
    BaseModel
        The result of the operation
    """
    rath = rath or current_rath.get()
    assert rath is not None, "No rath client provided and no rath client in current_rath context"
    x = await rath.aquery(
        operation.Meta.document,
        operation.Arguments(**variables).model_dump(by_alias=True),
    )
    return operation(**x.data)


def subscribe(operation: Type[TOperation], variables: Dict[str, Any], rath: Optional[Rath] = None) -> Iterator[TOperation]:
    """Synchronously subscribte to a subscription using rath


    This function will execute an operation using rath, retrieving
    the currentliy active rath client from the current_rath context
    if no rath client is provided.

    Parameters
    ----------
    operation : TurmsOperation
        The turms operation to execute
    variables : Dict[str, Any]
        The variables to use
    rath : Optional[Rath], optional
        The rath client, by default the current rath client

    Yields
    -------
    BaseModel
        The result of the operation
    """
    rath = rath or current_rath.get()
    assert rath is not None, "No rath client provided and no rath client in current_rath context"

    for event in rath.subscribe(
        operation.Meta.document,
        operation.Arguments(**variables).model_dump(by_alias=True),
    ):
        yield operation(**event.data)


async def asubscribe(operation: Type[TOperation], variables: Dict[str, Any], rath: Optional[Rath] = None) -> AsyncIterator[TOperation]:
    """Asynchronously subscribte to a subscription using rath


    This function will execute an operation using rath, retrieving
    the currentliy active rath client from the current_rath context
    if no rath client is provided.

    Parameters
    ----------
    operation : TurmsOperation
        The turms operation to execute
    variables : Dict[str, Any]
        The variables to use
    rath : Optional[Rath], optional
        The rath client, by default the current rath client

    yields
    -------
    BaseModel
        The result of the operation
    """
    rath = rath or current_rath.get()
    assert rath is not None, "No rath client provided and no rath client in current_rath context"

    async for event in rath.asubscribe(
        operation.Meta.document,
        operation.Arguments(**variables).model_dump(by_alias=True),
    ):
        yield operation(**event.data)



import json
from typing import Dict, Any, Union
from pydantic import BaseModel
from rath.operation import Operation
from rath.links.parsing import ParsingLink


ValidNestedTypes = Union[BaseModel, Dict[str, Any], list[Any], Any]


def parse_variables(
    variables: Dict[str, Any],
    by_alias: bool = True,
) -> ValidNestedTypes:
    """Parse Variables

    Parse vaiables converts any pydantic models in the variables dict to dicts
    by calling their .json() method with by_alias=True

    Args:
        variables (Dict): variables to parse
        by_alias (bool, optional): whether to use the alias names. Defaults to True.

    Returns:
        Dict: the parsed variables
    """

    def recurse_extract(obj: ValidNestedTypes) -> ValidNestedTypes:
        """
        recursively traverse obj, doing a deepcopy, but
        replacing any  pydantic BaseModels with their dict representation
        """

        if isinstance(obj, list):
            nulled_list: list[Any] = []
            for key, value in enumerate(obj):  # type: ignore
                value = recurse_extract(value)
                nulled_list.append(value)
            return nulled_list
        elif isinstance(obj, dict):
            nulled_obj = {}
            for key, value in obj.items():  # type: ignore
                value = recurse_extract(value)  # type: ignore
                nulled_obj[key] = value
            return nulled_obj  # type: ignore
        elif isinstance(obj, BaseModel):
            return json.loads(obj.model_dump_json(by_alias=by_alias))
        else:
            # base case: pass through unchanged
            return obj

    dicted_variables = recurse_extract(variables)

    return dicted_variables


class DictingLink(ParsingLink):
    """Dicting Link is a link that converts pydantic models to dicts.
    It traversed the variables dict, and converts any (nested) pydantic models to dicts
    by callind their .json() method."""

    by_alias: bool = True
    """Converts pydantic models to dicts by calling their .json() method with by_alias=True"""

    async def aparse(self, operation: Operation) -> Operation:
        """Dicts the variables

        Converts any pydantic models in the variables dict to dicts
        by calling their .json() method with by_alias=True

        Parameters
        ----------
        operation : Operation
            The operation to parse

        Returns
        -------
        Operation
            The parsed operation
        """

        shrinked_variables = parse_variables(operation.variables, by_alias=self.by_alias)
        operation.variables.update(shrinked_variables)
        return operation



import asyncio
from typing import AsyncIterator

from rath.links.base import ContinuationLink
from rath.operation import GraphQLResult, Operation
from rath.errors import NotComposedError


async def just_print(operation: Operation) -> None:
    """A simple logging function that prints the operation"""
    print(operation)


class TimeoutLink(ContinuationLink):
    """The Log link is a logging
    link that logs the operation to the console.
    links.
    """

    timeout: float = 20
    """The timeout for the operation in seconds"""

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """Executes an operation against the link

        This link simply forwards the operation to the next link.
        after logging it.

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Yields
        ------
        GraphQLResult
            The result of the operation
        """
        if not self.next:
            raise NotComposedError("No next link set")

        if not self.timeout:
            async for result in self.next.aexecute(operation):
                yield result

        else:
            aiter = self.next.aexecute(operation).__aiter__()
            try:
                while True:
                    # Wait for the next item from the async iterator
                    try:
                        result = await asyncio.wait_for(aiter.__anext__(), timeout=self.timeout)
                        yield result
                    except StopAsyncIteration:
                        break
            except asyncio.TimeoutError as e:
                raise asyncio.TimeoutError(f"Operation timed out after {self.timeout} seconds") from e
            except Exception as e:
                raise e



from rath.errors import NotComposedError
from rath.links.base import ContinuationLink
from rath.operation import GraphQLResult, Operation
from typing import AsyncIterator

import io
import aiohttp
from typing import AsyncGenerator
from pydantic import Field

from typing import Any, Dict, Tuple, Type, Union, List

FILE_CLASSES = (
    io.IOBase,
    aiohttp.StreamReader,
    AsyncGenerator,
)


ValidNestedTypes = Union[Dict[str, Any], List[Any], Any]


def parse_nested_files(
    variables: Dict[str, Any],
    file_classes: Tuple[Type[Any], ...] = FILE_CLASSES,
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Parse nested files

    Parameters
    ----------
    variables : Dict
        The variables to parse
    file_classes : Tuple[Type[Any], ...], optional
        File-like classes to extract, by default FILE_CLASSES

    Returns
    -------
    Tuple[Dict, Dict]
        The parsed variables and the extracted files
    """

    files = {}

    def recurse_extract(path: str, obj: ValidNestedTypes) -> ValidNestedTypes:
        """
        recursively traverse obj, doing a deepcopy, but
        replacing any file-like objects with nulls and
        shunting the originals off to the side.
        """
        nonlocal files

        if isinstance(obj, list):
            nulled_list = []
            for index, value in enumerate(obj):  # type: ignore
                value = recurse_extract(f"{path}.{str(index)}", value)
                nulled_list.append(value)  # type: ignore
            return nulled_list  # type: ignore
        elif isinstance(obj, dict):
            nulled_obj = {}
            for key, value in obj.items():  # type: ignore
                value = recurse_extract(f"{path}.{key}", value)  # type: ignore
                nulled_obj[key] = value
            return nulled_obj  # type: ignore
        elif isinstance(obj, file_classes):
            # extract obj from its parent and put it into files instead.
            files[path] = obj
            return None

        elif hasattr(obj, "__file__"):
            # extract obj from its parent and put it into files instead.
            files[path] = obj.value
            return None

        else:
            # base case: pass through unchanged
            return obj

    nulled_variables = recurse_extract("variables", variables)
    assert isinstance(nulled_variables, dict), "variables must be a dict"

    return nulled_variables, files  # type: ignore


class FileExtraction(ContinuationLink):
    """FileExtraction Link is a link that extracts file-like objects from the variables dict.
    It traverses the variables dict, and extracts any (nested) file-like objects to the context.files dict.

    These can then be used by the FileUploadLink to upload the files to a remote server.
    or used through the multipart/form-data encoding in the terminating link (if supported).
    """

    file_classes: Tuple[Type[Any], ...] = Field(default=FILE_CLASSES)

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """Extracts files from the variables dict

        Extracts any file-like objects from the variables dict and puts them into the context.files dict.
        The variables dict is then updated to replace the file-like objects with nulls.
        """
        if not self.next:
            raise NotComposedError("FileExtractionLink must be composed with another link")

        operation.variables, operation.context.files = parse_nested_files(operation.variables, file_classes=self.file_classes)

        async for result in self.next.aexecute(operation):
            yield result



from ssl import SSLContext
from typing import (
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    Literal,
    Optional,
    Any,
    Self,
    Type,
    TypedDict,
    NotRequired,
)
from graphql import OperationType
from pydantic import Field
import websockets
import json
import asyncio
import logging
import ssl
import certifi
from rath.links.errors import LinkNotConnectedError, TerminatingLinkError

from rath.operation import (
    GraphQLException,
    GraphQLResult,
    Operation,
    SubscriptionDisconnect,
)
from rath.links.base import AsyncTerminatingLink


logger = logging.getLogger(__name__)


GQL_WS_SUBPROTOCOL = "graphql-ws"

# all the message types
GQL_CONNECTION_INIT = "connection_init"
GQL_START = "start"
GQL_STOP = "stop"
GQL_CONNECTION_TERMINATE = "connection_terminate"
GQL_CONNECTION_ERROR = "connection_error"
GQL_CONNECTION_ACK = "connection_ack"
GQL_DATA = "data"
GQL_ERROR = "error"
GQL_COMPLETE = "complete"
GQL_CONNECTION_KEEP_ALIVE = "ka"

GQL_PING = "ping"
GQL_PONG = "pong"

WEBSOCKET_DEAD = "websocket_dead"
WEBSOCKET_CANCELLED = "websocket_cancelled"


class CorrectableConnectionFail(TerminatingLinkError):
    """A CorrectableConnectionFail is raised when a connection fails, but can be recovered from"""

    pass


class DefiniteConnectionFail(TerminatingLinkError):
    """A DefiniteConnectionFail is raised when a connection fails, and cannot be recovered from"""

    pass


class InvalidPayload(TerminatingLinkError):
    """A InvalidPayload is raised when a invalid payload is received"""

    pass


InitialConnectPayload = Dict[str, Any]
PongPayload = Dict[str, Any]


class TransportMessage(TypedDict):
    """A TransportMessage is a message that is sent over the transport layer.

    It is a dictionary that contains the type of the message and the payload.
    The payload can be any type of data, but it is usually a dictionary.

    """

    type: Literal[
        "connection_init",
        "connection_ack",
        "connection_terminate",
        "connection_error",
        "websocket_cancelled",
        "connection_keep_alive",
        "ka",
        "websocket_dead",
        "ping",
        "pong",
        "data",
        "error",
        "complete",
        "websocket_dead",
    ]
    payload: NotRequired[Any]
    error: NotRequired[Exception]


class GraphQLWSLink(AsyncTerminatingLink):
    """GraphQLWSLink is a terminating link that sends operations over websockets using
      websockets via the graphql-ws protocol. This is a
      new standard protocol, and should not be used for new projects.

    This is a terminating link, so it should be the last link in the chain.
    This is a stateful link, keeing a connection open and sending messages over it.

    """

    ws_endpoint_url: str
    """ The endpoint url to connect to """
    allow_reconnect: bool = True
    """ Should the websocket try to reconnect if it fails """
    time_between_retries: float = 4
    """ The sleep time between retries """
    max_retries: int = 3
    """ The maximum amount of retries before giving up """
    ssl_context: SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where())
    )

    on_connect: Optional[Callable[[InitialConnectPayload], Awaitable[None]]] = Field(
        exclude=True, default=None
    )
    """ A function that is called before the connection is established. If an exception is raised, the connection is not established. Return is ignored."""

    on_pong: Optional[Callable[[PongPayload], Awaitable[None]]] = Field(
        exclude=True, default=None
    )
    """ A function that is called before a pong is received. If an exception is raised, the connection is not established. Return is ignored."""
    heartbeat_interval_ms: Optional[int] = None
    """ The heartbeat interval in milliseconds (None means no heartbeats are 
    being send) """

    _connection_lock: Optional[asyncio.Lock] = None
    _connected: bool = False
    _alive: bool = False
    _send_queue: Optional[asyncio.Queue[str]] = None
    _connection_task: Optional[asyncio.Task[None]] = None
    _ongoing_subscriptions: Optional[Dict[str, asyncio.Queue[TransportMessage]]] = None

    async def aforward(self, message: str) -> None:
        """Forward a message to the server

        Puts the message in the send queue. If the queue is full, this will block until
        the queue is not full anymore.

        Parameters
        ----------
        message : str
            The message to send
        """
        if not self._send_queue:
            raise LinkNotConnectedError("Link is not connected")

        await self._send_queue.put(message)

    async def __aenter__(self) -> Self:
        """Enter the link, and initialize the connection"""

        self._ongoing_subscriptions = {}
        self._connection_lock = asyncio.Lock()
        self._send_queue = asyncio.Queue()
        return self

    async def aconnect(self, operation: Operation) -> None:
        """Connect to the server

        Parameters
        ----------
        operation : Operation
            The operation that is used for the first time to connect to the server
            can be used to send an initial payload
        """

        logger.info("Connecting Websockets")
        initial_connection_future = asyncio.get_running_loop().create_future()
        self._connection_task = asyncio.create_task(
            self.websocket_loop(operation, initial_connection_future)
        )
        await initial_connection_future

    async def adisconnect(self) -> None:
        """Disconnect from the server"""
        if self._connection_task:
            self._connection_task.cancel()

            try:
                await self._connection_task
            except asyncio.CancelledError:
                logger.info("Websocket Transport succesfully disconnected")

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        traceback: Optional[Any],
    ) -> None:
        """Exit the link, and disconnect from the server"""
        await self.adisconnect()

    async def abuild_url(self, initiating_operation: Operation) -> str:
        """Builds the url to connect to

        This can be overwritten to add additional parameters to the url
        like authentication tokens

        Parameters
        ----------
        initiating_operation : Operation
            The operation that is used for the first time to connect to the server

        Returns
        -------
        str
            The url to connect to
        """
        return self.ws_endpoint_url

    async def websocket_loop(
        self,
        initiating_operation: Operation,
        initial_connection_future: asyncio.Future[bool],
        retry: int = 0,
    ) -> None:
        """The main websocket loop

        This is the main loop that handles the websocket connection.
        It handles all the sending and receiving of messages.
        """
        send_task = None
        receive_task = None
        try:
            try:
                url = await self.abuild_url(initiating_operation)
                async with websockets.connect(  # type: ignore
                    url,
                    subprotocols=[
                        GQL_WS_SUBPROTOCOL,  # type: ignore
                    ],
                    ssl=self.ssl_context if url.startswith("wss") else None,
                ) as client:  # type: ignore
                    logger.info("Websocket successfully connected")

                    send_task = asyncio.create_task(
                        self.sending(
                            client,
                            initiating_operation,
                        )
                    )
                    receive_task = asyncio.create_task(
                        self.receiving(client, initial_connection_future)
                    )

                    self._alive = True
                    done, pending = await asyncio.wait(
                        (send_task, receive_task),
                        return_when=asyncio.FIRST_EXCEPTION,
                    )
                    self._alive = False

                    for task in pending:
                        task.cancel()

                    for task in done:
                        exception = task.exception()
                        if exception:
                            raise exception
                        else:
                            raise CorrectableConnectionFail(
                                f"Websocket connection closed without exception: This is unexpected behaviours. Results ist {task.result()}"
                            )

            except Exception as e:
                logger.warning(
                    f"Websocket excepted. Trying to recover for the {retry + 1}/{self.max_retries} time",
                    exc_info=True,
                )
                raise CorrectableConnectionFail from e

        except CorrectableConnectionFail as e:
            logger.info(
                f"Trying to Recover from Exception {e} Reconnect is {self.allow_reconnect} Retry: {retry}"
            )
            if retry > self.max_retries or not self.allow_reconnect:
                logger.error("Max retries reached. Aborting")
                raise DefiniteConnectionFail("Exceeded Number of Retries")

            await asyncio.sleep(self.time_between_retries)
            logger.info("Retrying to connect")
            await self.broadcast(
                {"type": WEBSOCKET_DEAD, "error": e}, initial_connection_future
            )
            await self.websocket_loop(
                initiating_operation, initial_connection_future, retry=retry + 1
            )

        except DefiniteConnectionFail as e:
            logger.error("Websocket excepted closed definetely", exc_info=True)
            self.connection_dead = True
            raise e

        except asyncio.CancelledError as e:
            logger.info("Websocket got cancelled. Trying to shutdown graceully")
            if send_task and receive_task:
                send_task.cancel()
                receive_task.cancel()

                await asyncio.gather(
                    send_task, receive_task, return_exceptions=True
                )  # wait for the tasks to finish
            raise e

        except Exception as e:
            logger.error("Websocket excepted", exc_info=True)
            self.connection_dead = True
            raise e

    async def sending(self, client: Any, initiating_operation: Operation) -> None:
        """The sending task

        This method is the sending task. It will send messages to the websocket
        as they are put into the send queue.  It should not be called manually.
        but is called automatically by the websocket loop.


        Parameters
        ----------
        client : websocket.Client
            The websockets client
        initiating_operation : Operation
            The initiating operation

        """

        payload: Dict[str, Any] = {
            "type": GQL_CONNECTION_INIT,
            "payload": initiating_operation.context.initial_payload,
        }
        await client.send(json.dumps(payload))

        try:
            while True:
                if not self._send_queue:
                    raise LinkNotConnectedError("Link is not connected")

                message = await self._send_queue.get()
                logger.debug("GraphQL Websocket: >>>>>> " + message)
                await client.send(message)
                self._send_queue.task_done()
        except asyncio.CancelledError as e:
            logger.debug("Sending Task sucessfully Cancelled")  #
            raise e

    async def receiving(
        self, client: Any, initial_connection_future: asyncio.Future[bool]
    ) -> None:
        """The receiving task

        This method is the receiving task. It will receive messages from the websocket
        and broadcast them to the subscriptions. It should not be called manually.
        but is called automatically by the websocket loop.

        Parameters
        ----------
        client : websocket.Client
            The websockets client
        connection_future : asyncio.Future
            The future to set when the connection is established

        """
        try:
            async for message in client:
                logger.debug("GraphQL Websocket: <<<<<<< " + message)
                try:
                    message = json.loads(message)
                    await self.broadcast(message, initial_connection_future)
                except json.JSONDecodeError as err:
                    logger.warning(
                        "Ignoring. Server sent invalid JSON data: %s \n %s",
                        message,
                        err,
                    )
        except Exception as e:
            logger.warning("Websocket excepted. Trying to recover", exc_info=True)
            raise e

    async def broadcast(
        self, message: TransportMessage, initial_connection_future: asyncio.Future[bool]
    ) -> None:
        """Broadcasts a message to all subscriptions"""
        type = message["type"]

        if type == GQL_CONNECTION_ACK:
            if self.on_connect:
                await self.on_connect(message.get("payload", {}))
            initial_connection_future.set_result(True)
            return

        if type == GQL_PING:
            if self.on_pong:
                await self.on_pong(message.get("payload", {}))

            payload = message.get("payload", {})
            await self.aforward(json.dumps({"type": GQL_PONG, "payload": payload}))

        if type == GQL_CONNECTION_KEEP_ALIVE:
            return

        if type == WEBSOCKET_DEAD:
            # notify all subscriptipns that the websocket is dead
            if not self._ongoing_subscriptions:
                self._ongoing_subscriptions = {}

            for subscription in self._ongoing_subscriptions.values():
                await subscription.put(message)
            return

        if type in [GQL_DATA, GQL_COMPLETE, GQL_ERROR]:
            if not self._ongoing_subscriptions:
                self._ongoing_subscriptions = {}

            if "id" not in message:
                raise InvalidPayload(f"Protocol Violation. Expected 'id' in {message}")

            id = message["id"]
            assert id in self._ongoing_subscriptions, (
                "Received Result for subscription that is no longer or was never active"
            )
            await self._ongoing_subscriptions[id].put(message)

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """Executes an operation against the link

        This link will send the operation to the websocket, and then
        wait for the result.

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Yields
        ------
        GraphQLResult
            The result of the operation
        """
        if not self._connection_lock:
            raise LinkNotConnectedError("Link is not connected")

        async with self._connection_lock:
            if self._connection_task is None or self._connection_task.done():
                # we need to start a new connection
                await self.aconnect(operation)

        assert operation.node.operation == OperationType.SUBSCRIPTION, (
            "Operation is not a subscription"
        )
        assert not operation.context.files, "We cannot send files through websockets"

        id = operation.id
        subscribe_queue: asyncio.Queue[TransportMessage] = asyncio.Queue()  # type: ignore
        if not self._ongoing_subscriptions:
            self._ongoing_subscriptions = {}

        self._ongoing_subscriptions[id] = subscribe_queue

        send_payload: Dict[str, Any] = {
            "headers": operation.context.headers,
            "query": operation.document,
            "variables": operation.variables,
        }

        try:
            frame: Dict[str, Any] = {
                "id": id,
                "type": GQL_START,
                "payload": send_payload,
            }
            await self.aforward(json.dumps(frame))
            logger.debug(f"Subcription started {operation}")

            while True:
                answer = await subscribe_queue.get()

                if answer["type"] == GQL_ERROR:
                    assert "payload" in answer, (
                        "Protocol Violation. Expected 'payload' in GQL_ERROR"
                    )
                    payload = answer["payload"]
                    if isinstance(payload, list):
                        error_list: list[Dict[str, Any]] = payload
                    else:
                        error_list: list[Dict[str, Any]] = [payload]

                    raise GraphQLException(
                        "\n".join([e["message"] for e in error_list])
                    )

                if answer["type"] == GQL_DATA:
                    assert "payload" in answer, (
                        "Protocol Violation. Expected 'payload' in GQL_DATA"
                    )
                    payload = answer["payload"]

                    if "errors" in payload:
                        raise GraphQLException(
                            "\n".join([e["message"] for e in payload["errors"]])
                        )

                    if "data" in payload:
                        yield GraphQLResult(data=payload["data"])
                        subscribe_queue.task_done()

                if answer["type"] == WEBSOCKET_DEAD:
                    raise SubscriptionDisconnect(
                        f"Subcription {id} failed propagating Error {operation}"
                    )

                if answer["type"] == GQL_COMPLETE:
                    logger.info(f"Subcription done {operation}")
                    return

        except asyncio.CancelledError as e:
            logger.debug(f"Subcription ended {operation}")
            await self.aforward(json.dumps({"id": id, "type": GQL_STOP}))
            raise e



import asyncio
from types import TracebackType
from typing import (
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    Optional,
    Self,
    Tuple,
    Type,
    Any,
)

from pydantic import Field, field_validator
from rath.links.base import AsyncTerminatingLink
from rath.links.testing.mock import AsyncMockResolver
from rath.operation import GraphQLResult, Operation
from graphql import FieldNode, OperationType
import uuid
from rath.links.errors import TerminatingLinkError


def target_from_node(node: FieldNode) -> str:
    """Extract the target aka. the aliased name from a FieldNode.

    Args:
        node (FieldNode): A GraphQL FieldNode.

    Returns:
        str: The target
    """
    return (
        node.alias.value if hasattr(node, "alias") and node.alias else node.name.value
    )


class ConfigurationError(TerminatingLinkError):
    """A Configuration Error"""


ResolverDict = Dict[str, Callable[[Operation], Awaitable[Dict[str, Any]]]]


class AsyncStatefulMockLink(AsyncTerminatingLink):
    """A Stateful Mocklink

    This is a mocklink that can be used to mock a GraphQL server.
    You need to pass resolvers to the constructor.

    In addition to AsyncMockLink this class also supports Subscription,
    and has internal state that needs to be handled.

    """

    timeout: float = 2

    query_resolver: ResolverDict = Field(default_factory=dict, exclude=True)
    mutation_resolver: ResolverDict = Field(default_factory=dict, exclude=True)
    subscription_resolver: ResolverDict = Field(default_factory=dict, exclude=True)
    resolver: ResolverDict = Field(default_factory=dict, exclude=True)

    _connection_lock: Optional[asyncio.Lock] = None
    _connected: bool = False
    _futures: Optional[Dict[str, asyncio.Future[GraphQLResult]]] = None
    _inqueue: Optional[asyncio.Queue[Tuple[Operation, str]]] = None
    _connection_task: Optional[asyncio.Task[None]] = None

    @field_validator(
        "query_resolver",
        "mutation_resolver",
        "subscription_resolver",
        "resolver",
        mode="before",
    )
    @classmethod
    def coerce_resolver(cls: Type["AsyncStatefulMockLink"], v: Any) -> Dict[str, Any]:
        """Coerce the resolver to a dict if it is a AsyncMockResolver"""
        if isinstance(v, AsyncMockResolver):
            return v.to_dict()
        return v

    async def __aenter__(self) -> Self:
        """Aenter the link and set up the internal state"""
        self._connection_lock = asyncio.Lock()
        return await super().__aenter__()

    async def aconnect(self, operation: Operation) -> None:
        """Connect the link and set up the internal state"""
        self._futures = {}
        self._inqueue = asyncio.Queue()
        _connection_future: asyncio.Future[bool] = asyncio.Future()  # type: ignore
        self._connection_task = asyncio.create_task(self.resolving(_connection_future))
        await _connection_future

    async def adisconnect(self) -> None:
        """Disconnect the link and clean up the internal state"""
        if self._connection_task:
            self._connection_task.cancel()
            self._connected = False

            try:
                await self._connection_task
            except asyncio.CancelledError:
                pass

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Aexit the link and clean up the internal state"""
        await self.adisconnect()

    async def resolving(self, connection_future: asyncio.Future[bool]) -> None:
        """A coroutine that resolves the incoming operations in
        an inifite Loop

        Raises:
            NotImplementedError: If the operation is not supported (aka not implemented)
        """
        connection_future.set_result(True)
        while True:
            if self._inqueue is None:
                raise ConfigurationError(
                    "You need to call aconnect before you can resolve operations"
                )

            operation, id = await self._inqueue.get()

            resolve_futures: list[Awaitable[Dict[str, Any]]] = []

            try:
                if operation.node.operation == OperationType.QUERY:
                    for op in operation.node.selection_set.selections:
                        if not isinstance(op, FieldNode):  # pragma: no cover
                            raise NotImplementedError("Only FieldNode are supported")
                        if op.name.value in self.query_resolver:
                            resolve_futures.append(
                                self.query_resolver[op.name.value](operation)
                            )
                        elif op.name.value in self.resolver:
                            resolve_futures.append(
                                self.resolver[op.name.value](operation)
                            )
                        else:
                            raise NotImplementedError(
                                f"Mocked Resolver for Query '{op.name.value}' not in resolvers"
                                f": {self.query_resolver}, {self.resolver}  for AsyncMockLink"
                            )

                if operation.node.operation == OperationType.MUTATION:
                    for op in operation.node.selection_set.selections:
                        if not isinstance(op, FieldNode):  # pragma: no cover
                            raise NotImplementedError("Only FieldNode are supported")
                        if op.name.value in self.mutation_resolver:
                            resolve_futures.append(
                                self.mutation_resolver[op.name.value](operation)
                            )
                        elif op.name.value in self.resolver:
                            resolve_futures.append(
                                self.resolver[op.name.value](operation)
                            )
                        else:
                            raise NotImplementedError(
                                f"Mocked Resolver for Query '{op.name.value}' not in resolvers:"
                                f"{self.mutation_resolver}, {self.resolver}  for AsyncMockLink"
                            )
                resolved = await asyncio.gather(*resolve_futures)

                if not self._futures:
                    self._futures = {}

                self._futures[id].set_result(
                    GraphQLResult(
                        data={
                            target_from_node(op): resolved[i]
                            for i, op in enumerate(
                                operation.node.selection_set.selections
                            )
                            if isinstance(op, FieldNode)
                        }
                    )
                )
            except AttributeError as t:
                raise ConfigurationError("No resolver for operation") from t

            except Exception as e:
                if not self._futures:
                    self._futures = {}

                self._futures[id].set_exception(e)

            self._inqueue.task_done()

    async def submit(self, o: Operation, id: str) -> None:
        """Submit an operation to the resolver"""
        if self._inqueue is None:
            raise ConfigurationError(
                "You need to call aconnect before you can submit operations"
            )
        await self._inqueue.put((o, id))

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """Executes an operation against the link

        This link will send the operation to the correct resolver and return the result.

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Yields
        ------
        GraphQLResult
            The result of the operation
        """
        if self._connection_lock is None:
            raise ConfigurationError(
                "You need to call __aenter__ before you can execute operations"
            )

        if not self._futures:
            self._futures = {}

        async with self._connection_lock:
            if not self._connected:
                await self.aconnect(operation)

        if (
            operation.node.operation == OperationType.QUERY
            or operation.node.operation == OperationType.MUTATION
        ):
            uniqueid = str(uuid.uuid4())
            self._futures[uniqueid] = asyncio.Future()
            await self.submit(operation, uniqueid)
            yield await asyncio.wait_for(self._futures[uniqueid], timeout=self.timeout)

        if operation.node.operation == OperationType.SUBSCRIPTION:
            assert len(operation.node.selection_set.selections) == 1, (
                "Only one Subscription at a time possible"
            )

            op = operation.node.selection_set.selections[0]
            if not isinstance(op, FieldNode):  # pragma: no cover
                raise NotImplementedError("Only FieldNode are supported")

            if op.name.value in self.subscription_resolver:
                iterator = self.subscription_resolver[op.name.value](operation)
            elif op.name.value in self.resolver:
                iterator = self.resolver[op.name.value](operation)
            else:
                raise NotImplementedError(
                    f"Mocked Resolver for Query '{op.name.value}' not in resolvers: {self.subscription_resolver}, {self.resolver}  for AsyncMockLink"
                )

            async for event in iterator:  # type: ignore
                yield GraphQLResult(data={target_from_node(op): event})

        else:
            raise NotImplementedError("Only subscription are mocked")



from typing import Any, Callable, AsyncIterator, Optional, Self, Type

from pydantic import Field
from rath.operation import Operation, GraphQLResult
from rath.links.base import TerminatingLink


class SplitLink(TerminatingLink):
    """SplitLink is a link that splits the operation into two paths. The operation is sent to the left path if the predicate returns true, otherwise to the right path.

    This is useful for exampole when implementing a cache link, where the operation is sent to the cache if it is a query, and to the network if it is a mutation.
    Or when subscriptions and queries are sent to different links (e.g. a websocket link and a http link).
    """

    left: TerminatingLink
    """The link that is used if the predicate returns true."""
    right: TerminatingLink
    """The link that is used if the predicate returns false."""
    split: Callable[[Operation], bool] = Field(exclude=True)
    """The function used to split the operation. This function should return a boolean. If true, the operation is sent to the left path, otherwise to the right path."""

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """Executes an operation against one of the links
        according to the predicate.

        Parameters
        ----------
        operation : Operation
            The operation to execute, will be to the predicate

        Yields
        ------
        GraphQLResult
            The result of the operation by one of the links
        """

        iterator = (
            self.left.aexecute(operation)
            if self.split(operation)
            else self.right.aexecute(operation)
        )

        async for res in iterator:
            yield res

    async def __aenter__(self) -> Self:
        """Enters the context manager of the link"""
        await self.left.__aenter__()
        await self.right.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        traceback: Optional[Any],
    ) -> None:
        """Exits the context manager of the link"""
        await self.left.__aexit__(exc_type, exc_val, traceback)
        await self.right.__aexit__(exc_type, exc_val, traceback)


def split(
    left: TerminatingLink, right: TerminatingLink, split: Callable[[Operation], bool]
) -> SplitLink:
    """
    Splits a Link into two paths. Acording to a predicate function. If predicate returns
    true, the operation is sent to the left path, otherwise to the right path.

    Parameters
    ----------
    left : TerminatingLink
        The link that is used if the predicate returns true.
    right : TerminatingLink
        The link that is used if the predicate returns false.
    split : Callable[[Operation], bool]
        The function used to split the operation. This function should return a boolean. If true,
        the operation is sent to the left path, otherwise to the right path.

    Returns
    -------
    SplitLink
        The split link that splits the operation according to the predicate.
    """
    return SplitLink(left=left, right=right, split=split)



from .auth import AuthTokenLink
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.backends import default_backend
from pydantic import field_validator
from cryptography.hazmat.primitives.asymmetric import rsa
from typing import Any, Callable, Awaitable, Dict
import jwt
from rath.operation import Operation


class SignLocalLink(AuthTokenLink):
    """SignLocalLink

    SignLocalLink is a link is a type of AuthTokenLink that
    crated a JWT token using a local private key,
    and sends it to the next link.

    """

    private_key: rsa.RSAPrivateKey

    @field_validator("private_key", mode="before")
    def must_be_valid_pem_key(cls, v: Any) -> rsa.RSAPrivateKey:
        """Validates that the private key is a valid PEM key"""
        try:
            try:
                if isinstance(v, str):
                    with open(v, "rb") as f:
                        v = f.read()
            except Exception as e:
                raise ValueError("Could not read file") from e

            key = serialization.load_pem_private_key(
                v, password=None, backend=default_backend()
            )

            # Check if it is an RSA key
            if not isinstance(key, rsa.RSAPrivateKey):
                raise ValueError("PEM key is not an RSA Private key.")

            return key
        except Exception as e:
            raise ValueError(f"Not a valid PEM key. {str(e)}")

    async def aretrieve_payload(self, operation: Operation) -> Dict[str, Any]:
        """Abstract method to retrieve the payload to sign

        Implement this method to retrieve the payload to sign.

        Parameters
        ----------
        operation : Operation
            The operation to sign

        Returns
        -------
        Dict[str, Any]
            The payload to sign by jwt with the private key

        """
        raise NotImplementedError("Please implement this method.")

    async def aload_token(self, operation: Operation) -> str:
        """Loads the token to send to the next link

        This method will call the aretrieve_payload method to retrieve the payload to sign,
        and then sign it with the private key. Please implement the aretrieve_payload method
        to retrieve the payload to sign.

        Parameters
        ----------
        operation : Operation
            The operation to sign

        Returns
        -------
        str
            The token to send to the next link

        """

        payload = await self.aretrieve_payload(operation)
        token = jwt.encode(payload, key=self.private_key, algorithm="RS256")  # type: ignore
        # rsa.RSAPrivateKey should be correctly typed in pyjwt
        return token

    async def arefresh_token(self, operation: Operation) -> str:
        """Refreshes the token

        This method is not implemented for SignLocalLink, as it is not needed.
        (Local certificates should not expire)
        """

        raise NotImplementedError


class ComposedSignTokenLink(SignLocalLink):
    """ComposedSignTokenLink

    ComposedSignTokenLink is a SignLocalLink that
    uses a payload retriever to retrieve the payload
    to sign, enables the user to use a custom payload
    retriever., without having to implement the entire
    SignLocalLink.
    """

    payload_retriever: Callable[[Operation], Awaitable[Dict[str, Any]]]
    """ The payload retriever to use to retrieve the payload to sign"""

    async def aretrieve_payload(self, operation: Operation) -> Dict[str, Any]:
        """Wraps the payload_retriever method"""
        return await self.payload_retriever(operation)
